#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
#include <X11/extensions/XShm.h>

#include <string>
#include <iostream>

//#include <GL/gl.h> 

// graphics output object
struct goo
{
   std::string title, lastError;
   unsigned width, height; // window
   unsigned renderWidth, renderHeight;
   bool initFullScreen;
   int frames;
   // xwindow
   enum wstate { windowed = 0, fullScreen = 1, toggleFullScreen = 2 };
   Display *dpy;
   int screen;
   Window win;
   GC gc;
   XVisualInfo *vi;
   // xvideo & xshm
   int xvPort;
   XvImage *xvImage;
   XShmSegmentInfo *shmInfo;
   int completionType;

   goo();

   bool init();
   bool checkVisuals();
   bool initXV();
   void createWindow();
   void windowState(wstate state=toggleFullScreen);
   void fixXOBug();
   void putImage();
   void loop();
   virtual void keypressed(int keysym) { }
   virtual void iteration() { putImage(); }
};

goo::goo()
{
   title=" ";
   screen=0;
   width=640;
   height=480;
   renderWidth=320;
   renderHeight=240;
   initFullScreen=true;
   shmInfo=NULL;
   xvImage=NULL;
   frames=0;
}

bool goo::init()
{
   dpy = XOpenDisplay(NULL);
   if (dpy == NULL) {
      lastError="Could not open X display";
      return false;
   }
   if (!checkVisuals()) {
      lastError="No suitable visual for graphics output object.";
      return false;
   }
   createWindow();
   return initXV();
}

bool goo::checkVisuals()
{
   XVisualInfo vinfo;
   int n;
   if (XMatchVisualInfo (dpy, screen, 16, TrueColor, &vinfo)) {
       /* 16 bit visual (fastest with TinyGL) */
   } else if (XMatchVisualInfo (dpy, screen, 24, TrueColor, &vinfo)) {
       /* 24 bit visual */
   } else if (XMatchVisualInfo (dpy, screen, 32, TrueColor, &vinfo)) {
       /* 32 bit visual */
   } else if (XMatchVisualInfo (dpy, screen, 8, PseudoColor, &vinfo)) {
       /* 8 bit visual */
   } else {
       /* no suitable visual */
       return false;
   }
   vi=XGetVisualInfo(dpy,VisualAllMask,&vinfo,&n);
   return true;
}

extern "C" XvImage  *XvShmCreateImage(Display*, XvPortID, int, char*, int, int, XShmSegmentInfo*);

bool goo::initXV()
{
   XvAdaptorInfo *ai;
   unsigned ign,num_adaptors;
   if (Success!=XvQueryExtension(dpy, &ign, &ign, &ign, &ign, &ign)) {
      lastError="XVideo Extension not found";
      return false;
   }
   if (Success!=XvQueryAdaptors(dpy, RootWindow(dpy, screen), &num_adaptors, &ai)) {
      lastError="Couldn't query XVideo Adaptors";
      return false;
   }
   if (num_adaptors<=0) {
      lastError="No XVideo Adaptor present";
      return false;
   }

   xvPort=ai[num_adaptors-1].base_id;

   XvFreeAdaptorInfo(ai);        

   printf("xv %i, depth %i, shm create image %i x %i\n", 
      xvPort, vi->depth, renderWidth, renderHeight);

   shmInfo=(XShmSegmentInfo*)malloc(sizeof(XShmSegmentInfo));
   xvImage = (XvImage*) XvShmCreateImage(
      dpy, xvPort, vi->depth==24 ? 3 : 0x36315652 // RGBA | RV16 //RGB565
      , 0, renderWidth, renderHeight, shmInfo
   );
   shmInfo->shmid = shmget(IPC_PRIVATE, xvImage->data_size, IPC_CREAT | 0777);
   shmInfo->shmaddr = xvImage->data = (char*) shmat(shmInfo->shmid, 0, 0);
   shmInfo->readOnly = False;

   if (vi->depth==24)
      for (int i = 0; i < xvImage->height; i++) {
         for (int j = 0; j < xvImage->width; j++) {
            xvImage->data[(xvImage->width*i + j)*4] = i;
            xvImage->data[(xvImage->width*i + j)*4+1] = j;
            xvImage->data[(xvImage->width*i + j)*4+2] = 255-i;
         }
      }
   else {
      int i;
      for (i = 0; i < xvImage->height; i++) {
         for (int j = 0; j < xvImage->width; j++) {
            unsigned short t2=(i>>3)|((j>>2)<<5)|(((255-i)>>3)<<11);
            unsigned short t;
            t=(t2&0x1f)>>2;
            t|=(t2&0x7e0)>>1;
            t|=(t2&0xf8)>>2;
            xvImage->data[(xvImage->width*i + j)*2] = t&0xff;
            xvImage->data[(xvImage->width*i + j)*2+1] = (t>>8)&0xff;
         }
      }
   }
   
   completionType=XShmGetEventBase(dpy) + ShmCompletion;

   if (!XShmAttach(dpy, shmInfo)) {
      lastError="Couldn't attach the shared memory to the X display";
      return false;
   }
   return true;
}

void goo::fixXOBug()
{
   int i;
   for (i = 0; i < xvImage->height; i++) {
      for (int j = 0; j < xvImage->width; j++) {
         unsigned short t2=xvImage->data[(xvImage->width*i + j)*2];
         t2|=xvImage->data[(xvImage->width*i + j)*2+1]<<8;
         unsigned short t;
         t=(t2&0x1f)>>2;
         t|=(t2&0x7e0)>>1;
         t|=(t2&0xf8)>>2;
         xvImage->data[(xvImage->width*i + j)*2] = t&0xff;
         xvImage->data[(xvImage->width*i + j)*2+1] = (t>>8)&0xff;
      }
   }
}

static Bool waitForNotify(Display *d, XEvent *e, char *arg) 
{
  return (e->type == MapNotify) && (e->xmap.window == (Window)arg); 
}

void goo::createWindow()
{
   Colormap cmap;
   XSizeHints hint;
   XSetWindowAttributes swa;

   if (initFullScreen)
   {
      unsigned depth;
      Window rootwnd;
      XGetGeometry(dpy, RootWindow(dpy, screen), 
          &rootwnd, // whatever
          &hint.x, &hint.y,
          &width, &height,
          &depth, &depth // whatever
      );
   }

   cmap = XCreateColormap(dpy, RootWindow(dpy, screen),
          vi->visual, AllocNone);

   swa.colormap = cmap;
   swa.border_pixel = 0;
   swa.event_mask = StructureNotifyMask;

   hint.x = hint.y = 0;
   hint.width = width;
   hint.height = height;
   hint.flags = PPosition | PSize;
   
   win = XCreateWindow(dpy, RootWindow(dpy, screen), 0, 0, width, height,
		      0, vi->depth, InputOutput, vi->visual, 
		      CWBorderPixel | CWEventMask | CWColormap, &swa
   );
   XSetStandardProperties (dpy, win, title.c_str(), title.c_str(), None, NULL, 0, &hint);
   XMapRaised(dpy, win);

   windowState(fullScreen);
  
   XEvent event;
   XIfEvent(dpy, &event, waitForNotify, (char*)win);
   XSelectInput(dpy, win, KeyPressMask | StructureNotifyMask | ExposureMask);
   
   XGCValues xgcv;
   gc = XCreateGC( dpy, win, 0L, &xgcv );

   //XSetForeground( dpy, gc, 0x008090a0L );
   XFillRectangle( dpy, win, gc, 0, 0, width, height );
}

void goo::windowState(wstate state)
{
   Atom wmState = XInternAtom(dpy, "_NET_WM_STATE", False);
   Atom fullScreen = XInternAtom(dpy, "_NET_WM_STATE_FULLSCREEN", False);

   XEvent xev;
   xev.xclient.type=ClientMessage;
   xev.xclient.serial = 0;
   xev.xclient.send_event=True;
   xev.xclient.window=RootWindow(dpy,screen);
   xev.xclient.message_type=wmState;
   xev.xclient.format=32;
   xev.xclient.data.l[0] = 1; //state;
   xev.xclient.data.l[1] = fullScreen;
   xev.xclient.data.l[2] = 0;

   XSendEvent(dpy, win, False,
              SubstructureRedirectMask | SubstructureNotifyMask,
              &xev
   ); 
}

static Bool waitForShmCompletion(Display *dpy, XEvent *event, char *arg) 
{
    goo *g=(goo *) arg;
    return (event->type == g->completionType) && 
        ( ((XShmCompletionEvent *)event)->drawable == g->win); 
}

void goo::putImage()
{
   frames++;
   XEvent event;
   XvShmPutImage(dpy, xvPort, win, gc, xvImage,
         0, 0, xvImage->width, xvImage->height,
         0, 0, width, height, True);

   XIfEvent(dpy, &event, waitForShmCompletion, (char*)this);
}

void goo::loop()
{
   XEvent xev;
   char buf[80];
   KeySym keysym;
   XComposeStatus status;
   bool done=false;
   while (!done) 
   {
      if (XPending(dpy) > 0) 
      {
         XNextEvent(dpy,&xev);
         switch(xev.type) {
            case KeyPress:
	            XLookupString((XKeyEvent *)&xev,buf,80,&keysym,&status);
	            windowState();
	            keypressed(keysym);
	            printf("keypressed %x\n",keysym);
	            if (keysym==0xff1b)
	               done=true;
               break;
            case ConfigureNotify:
            {
               int width,height;
               width = xev.xconfigure.width;
               height = xev.xconfigure.height;
	            printf("config window %i x %i\n",width,height);
               break;
            }
         }
      } else {
         iteration();
      }
   }
}

#include <GL/gl.h> 
#include <GL/oscontext.h> 
#include "ui.h"

struct gltest : goo
{
   void *fbuff;
   ostgl_context *cx;

   void glInit()
   {
      //fbuff=(void*)malloc(renderWidth*renderHeight*2);
      fbuff=(void*)xvImage->data;
      cx = ostgl_create_context(renderWidth, renderHeight, 16, &fbuff, 1);
      ::init();
      reshape(renderWidth, renderHeight);
   }
   ~gltest()
   {
      ostgl_delete_context(cx);
   }
   virtual void iteration()
   {
      idle();
   }
   virtual void keypressed(int keysym) 
   {
      key(keysym,(GLenum)0);
   }
};

gltest g;

void tkSwapBuffers(void)
{
   //if (g.vi->depth!=24)
   //   g.fixXOBug();
   g.putImage();
}


int ui_loop(int argc, char **argv, const char *name)
{
   g.title="Aplicativo de Teste";
   if (!g.init())
   {
      std::cout << g.lastError << "\n";
      return 0;
   }
   g.glInit();
   g.loop();
   std::cout << "frames=" << g.frames << "\n";
}

